# $Id: controller.py,v 1.34 2010/11/22 17:44:57 irees Exp $
import os
import re
import operator
import Queue
import collections
import time
import datetime
from PyQt4 import QtCore, QtGui

import emen2.clients.controllers
import emen2.clients.emdash.threads
import emen2.clients.emdash.config


#emen2.clients.controllers.EMEN2ClientController
class BaseController(QtCore.QThread):		

	# Update the record mapping, cached records, and recnames.
	signal_update_recs = QtCore.pyqtSignal(object, object, object)
	signal_update_comments = QtCore.pyqtSignal(object)

	# Changes in file status
	signal_update_setting = QtCore.pyqtSignal(unicode, unicode)
	signal_newfile = QtCore.pyqtSignal(unicode, object)
	signal_file_status = QtCore.pyqtSignal(unicode, object)
	signal_file_success = QtCore.pyqtSignal(unicode, object)
	signal_file_failure = QtCore.pyqtSignal(unicode, unicode)

	# Session controls
	signal_begin_session = QtCore.pyqtSignal(int)
	signal_end_session = QtCore.pyqtSignal()
	signal_newgrid = QtCore.pyqtSignal(int)

	# Login
	signal_getlogin = QtCore.pyqtSignal(unicode, unicode)
	signal_login = QtCore.pyqtSignal(unicode, unicode, unicode)
	signal_exception = QtCore.pyqtSignal(unicode)

	# Change a value in a record: recid, parameter, new value
	signal_set = QtCore.pyqtSignal(int, unicode, unicode)

	# Change a controller setting: parameter, new value
	signal_setting = QtCore.pyqtSignal(unicode, object)

	
	def __init__(self, gui=False, parent=None):
		QtCore.QThread.__init__(self, parent=parent)

		# Database
		self.db = emen2.clients.emdash.config.getdb()

		# Record IDs, record names, comments
		self.bdos = {}
		self.comments = {}
		self.recs = {}
		self.recnames = {}		
		self.recids = {}
				
		# Current microscope settings...		
		self.settings = {}
		self.settings['assess_image_quality'] = 5
		
		# Queue of currently un-uploaded files
		self.queue = Queue.Queue()

		self._setupworkers()



	###############################
	# Other Threads
	###############################

	def _setupworkers(self):
		# Starts when a new microscopy session is made
		self.clockthread = emen2.clients.emdash.threads.clockthread.ClockThread(parent=self)

		# FSPollThread starts whenever a directory is selected
		self.fspollthread = emen2.clients.emdash.threads.fspollthread.FSPollThread(parent=self) # queue=self.queue
		self.fspollthread.signal_newfile.connect(self.newfile, type=QtCore.Qt.QueuedConnection)
		
		# Starts when a grid has been selected or created
		self.uploadthread = emen2.clients.emdash.threads.uploadthread.UploadThread(parent=self, db=self.db, queue=self.queue)

		self.uploadthread.signal_file_status.connect(self.signal_file_status, type=QtCore.Qt.QueuedConnection)
		self.uploadthread.signal_file_success.connect(self.signal_file_success, type=QtCore.Qt.QueuedConnection)
		self.uploadthread.signal_file_failure.connect(self.signal_file_failure, type=QtCore.Qt.QueuedConnection)
		
		# Gracefully stop processing when session ends..
		# self.signal_end_session.connect(self.fspollthread.quit, type=QtCore.Qt.QueuedConnection)
		# self.signal_end_session.connect(self.clockthread.quit, type=QtCore.Qt.QueuedConnection)



	###############################
	# Session/Record ID management
	###############################

	def _getrecids(self):
		return filter(None, self.recids.values())


	def _reset(self, exclude=None):
		"""Reset the controller"""
		
		if exclude == None:
			exclude = []
		if not hasattr(exclude, "__iter__"):
			exclude = [exclude]
		
		exclude.append('microscope')
		backup = {}

		for i in exclude:
			if self.recids.get(i) != None:
				backup[i] = self.recids.get(i)
			
		self.recids = backup
		self.update_recs()
				


	def run(self):
		print "Waiting for queue to empty..."
		self.queue.put({"filename":None})
		self.uploadthread.wait()
		self.finished.emit()
		self.db.logout()
		print "Queue is empty; closing session..."

	

	@QtCore.pyqtSlot(str, str)
	def login(self, username=None, password=None):
		# self.options.username = username or self.options.username or raw_input("Username: ")
		# self.options.password = password or self.options.password or getpass.getpass("Password: ")

		self._reset()
		try:
			self.db._login(unicode(username), unicode(password))
			self.username, self.groups = self.db.checkcontext()
			user = self.db.getuser(self.username) or {}
			self.displayname = user.get('displayname', self.username)
			self.signal_login.emit(self.db._ctxid, self.username, self.displayname)						
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			return
			
			
			
	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def begin_no_session(self, ctxid, username, displayname):	
		# begin_session directly from login callback
		print "Begin Session"
		self.clockthread.start()		
		self.uploadthread.start()
		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(-1)
		
	

	@QtCore.pyqtSlot(object, object)
	def begin_session(self, editrecords, select_records):	
		# Find the correct microscope
		print "Begin Session"
		self.clockthread.start()		
		self.uploadthread.start()
		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(self.recids.get(emen2.clients.emdash.config.options.session_rectype))

			

		
	@QtCore.pyqtSlot()	
	def end_session(self, restart=True):
		self._reset()
		self.signal_end_session.emit()		
		self.clockthread.quit()
		self.fspollthread.quit()
		self.start()


		
	# Controller discovers a new file, does some processing, and rebroadcasts.	
	# @QtCore.pyqtSlot(unicode, object)	
	def newfile(self, filename, d=None):
		d = d or {}
		filename = unicode(filename)

		ext = os.path.splitext(filename)[-1]
		auto = False
		if ext in emen2.clients.emdash.config.autoqueue:
			auto = True

		if ext not in emen2.clients.emdash.config.autoqueue+emen2.clients.emdash.config.manualqueue:
			return

		d.update({'_recname': os.path.basename(filename), '_filename':filename, "_auto":auto})
		d.update(self.settings)
		print "New file: %s"%filename
		self.signal_newfile.emit(filename, d)




	###############################
	# Editing records / Comments
	###############################


	@QtCore.pyqtSlot(int, unicode, unicode)
	def set(self, recid, param, value):
		rec = self.db.putrecordvalues(recid, {unicode(param): unicode(value)})
		self.signal_file_status.emit(str(recid), rec)



	@QtCore.pyqtSlot(int, unicode)
	def addcomment(self, recid, text):
		try:
			self.db.addcomment(recid, unicode(text))
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			return

		self.update_comments()



	def _getcomments(self):
		return []
		

	def update_comments(self):
		self.comments = self._getcomments()
		self.signal_update_comments.emit(self.comments)




	def update_recs(self, newmapping=None, quiet=False):
		"""Update the current set of records."""

		newmapping = newmapping or {}
		if newmapping:
			self.recids.update(newmapping)

		rnget = self._getrecids()

		recs = self.db.getrecord(rnget)
		for rec in recs:
			self.recs[rec.get('recid')] = rec

		rn = self.db.renderview(rnget, None, "recname")
		self.recnames.update(rn)		

		if not quiet:
			self.signal_update_recs.emit(self.recids, self.recs, self.recnames)



	def set_setting(self, param, value):
		# print "Changing setting: %s -> %s"%(param, value)
		self.settings[str(param)] = str(value)
		self.signal_update_setting.emit(str(param), str(value))



	
	
	###############################
	# Some controller specific behaviors...
	###############################


	# Fold this into something else..
	def newmicrograph(self):
		id_micrograph = self.settings.get('id_micrograph','')
		matches = re.search('(\w+?)(\d+)', id_micrograph)
		if not matches:
			print "Error creating new micrograph: No Micrograph ID..."
			return

		matches = matches.groups()		
		newid = int(matches[1])+1
		newid = matches[0] + str(newid).zfill(len(matches[1]))
		
		data = {'_recname': id_micrograph, '_filename':id_micrograph, 'id_micrograph':id_micrograph, "_auto":True, 'rectype':'micrograph'}
		data.update(self.settings)

		# Emit the new file signal, and update the UI to show the new file micrograph ID
		self.signal_newfile.emit(id_micrograph, data)
		self.set_setting('id_micrograph', newid)
		
	
	
	@QtCore.pyqtSlot(object, object)
	def newfreezing(self, newrecords, selected_records):
		try:
			newrecords = self.db.putrecord(newrecords)
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise


	@QtCore.pyqtSlot(int)
	def selectmicroscopy(self, recid):
		self.begin_session([], {emen2.clients.emdash.config.options.session_rectype:recid})


	@QtCore.pyqtSlot(int)
	def selectgrid(self, recid):
		self.newgrid([], {'grid_imaging':recid})



	# ian: short-circuit signal since everything is converted to QStrings and blows up xmlrpc
	@QtCore.pyqtSlot(object, object)
	def newgrid(self, newrecords, selected_records):
		self._reset(exclude=emen2.clients.emdash.config.options.session_rectype)

		try:
			newrecords = self.db.putrecord(newrecords)
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise

		for rec in newrecords:
			selected_records[rec.get("rectype")] = rec.get("recid")

		self.update_recs(newmapping=selected_records)		

		# ian: todo. I changed to use page names, instead of record types.. but this may cause some problems.
		e = selected_records.get("grid_imaging")
		print "Grid Imaging Session: %s"%e
		if e > -1:
			self.signal_newgrid.emit(int(selected_records.get("grid_imaging")))




	@QtCore.pyqtSlot(object, object)
	def edit_records(self, editrecords, selected_records):
		recs = {}
		# xmlrpc requires keys to be strs
		for rec in editrecords:
			recs[str(rec['recid'])] = rec
		
		try:
			self.db.putrecordsvalues(recs)
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise

		self.update_recs()




class MicroscopeController(BaseController):
	
	def _getcomments(self):
		root = self.recids.get('microscope')
		if  root != None:
			recs = self.db.getchildren(self.recids["microscope"], 1, ["microscopy*"])
			coms = self.db.getcomments(self._getrecids()+recs)
			ret = collections.defaultdict(list) # backwards compat
			for i in coms:
				ret[i[0]].append(i[1:])
			return ret
			
			
	
	def begin_session(self, editrecords, select_records):		
		# Find the correct microscope
		t = self.db.gettime()
		microscopes = self.db.getrecord(self.db.getchildren(emen2.clients.emdash.config.options.microscope, 1, "microscope"))
		microscope = None
		for m in microscopes:
			# print m.get('date_start'), m.get('date_end')
			if m.get('date_start') <= t < m.get('date_end'):
				self.recids['microscope'] = m.get('recid')
							
		if not self.recids.get('microscope'):
			raise ValueError, "No microscope valid for today's date: %s... tried children of %s"%(t, emen2.clients.emdash.config.options.microscope)
		
		if select_records.get(emen2.clients.emdash.config.options.session_rectype):
			rec = self.db.getrecord(select_records.get(emen2.clients.emdash.config.options.ssession_rectype))

		else:
			rec = self.db.newrecord(emen2.clients.emdash.config.options.session_rectype, self.recids.get("microscope"))
			rec["date_start"] = self.db.gettime()
			rec["parents"] = [self.recids.get("microscope")]
			rec.update(editrecords[0])
			rec = self.db.putrecord(rec)

		self.recids[emen2.clients.emdash.config.options.session_rectype] = rec.get("recid")
		self.clockthread.start()		
		self.uploadthread.start()

		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(self.recids[emen2.clients.emdash.config.options.session_rectype])

		
		
	def end_session(self, restart=True):
		if self.recids.get(emen2.clients.emdash.config.options.session_rectype) != None:
			try:
				rec = self.db.getrecord(self.recids.get(emen2.clients.emdash.config.options.session_rectype))
				rec["date_end"] = self.db.gettime()
				rec = self.db.putrecord(rec)
			except:
				print "Error ending session record"

		self._reset()
		self.signal_end_session.emit()
		
		self.clockthread.quit()
		self.fspollthread.quit()
		self.start()
	


__version__ = "$Revision: 1.34 $".split(":")[1][:-1].strip()
