# $Id: downloadresource.py,v 1.55 2010/11/21 22:41:45 irees Exp $
import urllib2
import re
import time
import collections
import re
import os
import pickle
import traceback
import time
import random
import tarfile
import StringIO
import cStringIO

import twisted.web.static

# emen2 imports
import emen2.web.resources.emen2resource
import emen2.db.exceptions
from emen2.web import routing
from emen2.web import responsecodes

import emen2.util.jsonutil
import emen2.web.thumbs
import emen2.db.config
g = emen2.db.config.g()



class DownloadResource(emen2.web.resources.emen2resource.EMEN2Resource):

	resourcename = 'download'
	defaultType = 'application/octet-stream'

	def _action(self, path=None, postpath=None, args=None, ctxid=None, host=None, method='GET', db=None, request=None):
		"""thread worker to get file paths from db; hand back to resource to send """

		# The args will work like a db.query -- take out size, format first.
		size = args.pop("size", None)
		format = args.pop('format', 'jpg')
		q = args.get('q')

		# Any elements in the path will be treated as bdokey=..
		postpath = postpath or ['']
		# The order is important here -- check the args before path.
		bids = args.get('bids', []) or filter(bool, postpath[0].split(","))
		username = None
		bdo = None

		if bids and not hasattr(bids, "__iter__"):
			bids = [bids]

		with db:
			db._setContext(ctxid,host)
			username = db._getctx().username

			# Handle temporary files specially
			if postpath[0] == "tmp":
				if postpath[1].partition("-")[0] != ctxid:
					msg =  "Permission denied for %s"%path
					g.log.msg('LOG_INFO', msg)
					raise IOError, msg

				filename = "-".join(postpath[1:])
				filepath = os.path.join(g.paths.TMPPATH, filename)
				bdos = [{'filename':filename, 'filepath':filepath}]

			# Query for BDOs
			elif q:
				bdos = db.getbinary(q=q)

			else:
				bdos = db.getbinary(bids)


		files = {}
		for bdo in bdos:
			filename = bdo.get("filename")
			filepath = bdo.get("filepath")

			# If we're looking for a particular size or format..
			if size:
				filename = "%s.%s.%s"%(filename, size, format)
				filepath = os.path.join(g.paths.TILEPATH, "%s.%s.%s"%(bdo.name.replace(":","."), size, format))

				# If the size/format doens't exist, try to create it.
				if not os.access(filepath, os.F_OK):
					try:
						emen2.web.thumbs.run_from_bdo(bdo, wait=True)
					except Exception, inst:
						pass

			files[filepath] = filename
			if not os.access(filepath, os.F_OK):
				msg = "Could not access: %s"%filepath
				g.log.msg('LOG_INFO', msg)
				raise IOError, msg

		if len(files) == 0:
			raise IOError, "No files found"

		return files, {'X-Username':username, 'X-Resource': 'upload'}



	def _cb(self, result, request, ctxid=None, t=0):
		# This resource returns files or a file stream, so it's a little more complicated
		files, content_headers = result
		self.alwaysCreate = False
		request.setResponseCode(200)

		if len(files) > 1:
			return self._transfer_tar(files, request)

		return self._transfer_single(files, request)



	def _transfer_single(self, files, request):
		filepath, filename = files.items()[0]
		mimetype, encoding = twisted.web.static.getTypeAndEncoding(filename, self.contentTypes, self.contentEncodings, self.defaultType)

		# If we're saving a gzip, we'll let the browser expand it by setting encoding:gzip.
		if filename[-3:] == ".gz":
			filename = filename[:-3]
			encoding = "gzip"

		fsize = os.stat(filepath).st_size
		f = open(filepath)

		if request.postpath[-1] == "save":
			request.setHeader('Content-Disposition', 'attachment; filename=%s'%filename.encode('utf-8'))

		request.setHeader('Content-Length', str(fsize))
		request.setHeader('Content-Type', mimetype)
		request.setHeader('Content-Encoding', encoding)


		a = twisted.web.static.NoRangeStaticProducer(request, f)
		a.start()


	def _transfer_tar(self, files, request):
		request.setHeader('Content-Disposition', 'attachment; filename=archive.tar')
		request.setHeader('Content-Type', 'application/x-tar')
		request.setHeader('Content-Encoding', self.defaultType)

		a = twisted.web.static.NoRangeStaticProducer(request, TarPipe(files))
		a.start()





class TarPipe(object):
	"""This class implements a compression pipe suitable for asynchronous
	process."""

	# ian: todo: stream in added gz files, and write compressed tar output. Basically repackaged a bunch of .gz's to a .tar.gz with non-gz's inside.
	def __init__(self, files={}):
		self.pos = 0
		self.files = files

		# StringIO.StringIO.__init__(self)
		self.cbuffer = cStringIO.StringIO()
		# self.cbuffer = ''
		self.tarfile = tarfile.open(mode='w|', fileobj=self)


	def close(self):
		pass


	def _addnextfile(self):
		if not self.files:
			return

		key = self.files.keys()[0]
		filename = self.files.pop(key)

		self.pos = 0
		self.cbuffer.seek(self.pos)
		self.cbuffer.truncate(0)

		self.tarfile.add(key, arcname=filename)

		# print "Added %s / %s.. buffer size is %s. %s files left"%(key, filename, 0, len(self.files))

		if len(self.files) == 0:
			# print "Closing tarfile"
			self.tarfile.close()

		self.cbuffer.seek(self.pos)



	def write(self, data):
		self.cbuffer.write(data)


	def read(self, size=65536):
		data = self.cbuffer.read(size) # [self.pos:self.pos+size]

		if len(data) == 0:
			self._addnextfile()
			data = self.cbuffer.read(size) #[self.pos:self.pos+size]

		self.pos += len(data)
		# print "set pos to %s"%self.pos

		return data



__version__ = "$Revision: 1.55 $".split(":")[1][:-1].strip()
