# $Id: export.py,v 1.12 2010/11/15 04:23:25 irees Exp $
import time
import sys
import os
import operator

import xml.etree.ElementTree
etree = xml.etree.ElementTree

import xml.dom.minidom
md = xml.dom.minidom

# Standard View imports
import emen2.db.config
g = emen2.db.config.g()
from emen2.web.view import View
###



@View.register
class ExportEMDB(View):

	@View.add_matcher(r'^/export/emdb/(?P<recid>\d+)/$')
	def init(self, recid=None):
		self.template="/pages/export.emdb"
		self.title = "Export to EMDB"
		self.ctxt["recid"]=int(recid)

		exported = harvest(int(recid), db=self.db)

		self.ctxt["exported"] = exported



nodemap = {
	"root": ["emdEntry"],

	"emdEntry": ["deposition", "map", "supplement", "sample", "experiment", "processing", ],

	"deposition": ["mapReleaseDate", "fittedPDBEntryIDs", "primaryReference"],
		"primaryReference": ["journalArticle"],
			"journalArticle": ["authors", "articleTitle", "journal", "volume", "firstPage", "lastPage", "year", "externalReference"],

	"map": ["file", "dataType", "dimensions", "origin", "limit", "cell", "axisOrder", "statistics", "spaceGroupNumber", "details", "pixelSpacing", "contourLevel", "annotationDetails"],
		"dimensions": ["numColumns", "numRows", "numSections"],
		"origin": ["originCol", "originRow", "originSec"],
		"limit": ["limitCol", "limitRow", "limitSec"],
		"cell": ["cellA", "cellB", "cellC", "cellAlpha", "cellBeta", "cellGamma"],
		"axisOrder": ["axisOrderFast", "axisOrderMedium", "axisOrderSlow"],
		"statistics": ["minimum", "maximum", "average", "std"],
		"pixelSpacing": ["pixelX", "pixelY", "pixelZ"],

	"supplement": ["maskSet", "sliceSet", "figureSet", "fscSet"],

	"sample": ["numComponents", "aggregationState", "name", "compDegree", "molWtTheo", "sampleComponentList"],
		"sampleComponentList": ["sampleComponent"],
			"sampleComponent": ["protein"],
				"protein": ["refInterpro", "mutantFlag", "oligomericDetails", "refGo", "numCopies", "natSource", "engSource", "synName", "entry", "sciName", "molWtTheo"],
					"natSource": ["cell"],
					"engSource": ["vector", "expSystem"],


	"experiment": ["vitrification", "imaging", "imageScans", "fitting", "samplePreparation"],
		"vitrification": ["method", "cryogenName", "humidity", "instrument", "temperature"],
		"imaging": ["electronSource", "electronDose", "imagingMode", "nominalDefocusMin", "nominalDefocusMax", "illuminationMode", "temperatureMax", "specimenHolder", "details", "detector", "nominalCs", "temperature", "microscope", "date", "specimenHolderModel", "acceleratingVoltage", "nominalMagnification"],
		"imageScans": ["scanner", "numDigitalImages", "samplingSize"],
		"samplePreparation": ["buffer"],
			"buffer": ["details"],

	"processing": ["reconstruction", "emDataSet"],
		"reconstruction": ["algorithm", "software", "ctfCorrection", "resolutionByAuthor", "resolutionMethod"],
		"emDataSet": ["singleParticle"],
			"singleParticle": ["numProjections"]
}

parammap = {
	"name_specimen": "sample/name",
	"mass_specimen": "sample/molWtTheo"
}



def harvest(recid, db):


	#allnodes = set(reduce(operator.concat, nodemap.values()+nodemap.keys()))

	root = recid

	#############

	# Basic Database Structure

	root = db.getrecord(root)

	subproject = db.getrecord(db.getparents(root.recid, rectype=["subproject"])).pop()
	project = db.getrecord(db.getparents(root.recid, rectype=["project"], recurse=2)).pop()




	####################
	# XML: emdEntry (root)
	e = etree.ElementTree(etree.Element("emdEntry", accessCode="ABC-123"))
	for i in nodemap["emdEntry"]:
		e2 = etree.Element(i)
		e.find("/").append(e2)



	####################
	# XML: emdEntry / deposition
	e_deposition = e.find('/deposition/')

	e_mapReleaseDate = etree.Element("mapReleaseDate")
	e_mapReleaseDate.text = time.strftime("%m-%d-%Y")
	e_deposition.append(e_mapReleaseDate)

	e_replaceExistingEntry = etree.Element("replaceExistingEntry")
	e_replaceExistingEntry.text = "false"
	e_deposition.append(e_replaceExistingEntry)


	# XML: emdEntry / deposition / primaryReference
	publications = db.getrecord(db.getchildren(subproject.recid, rectype=["publication"]))
	for pub in publications:
		if pub.get('pmid'):
			e_primaryReference = etree.Element("primaryReference")
			e_externalReference = etree.Element("externalReference", type="pubmed")
			e_externalReference.text = str(pub.get('pmid'))
			e_primaryReference.append(e_externalReference)
			e_deposition.append(e_primaryReference)



	####################
	# XML: emdEntry / map
	e_map = e.find("/map/")

	mapbdo = db.getbinary(root.get("reconstruction_map"))
	mapinfo = os.stat(mapbdo.get("filepath"))
	sizeKb = str(mapinfo.st_size / 1024)

	# mapemdata = EMAN2.EMData()
	# mapemdata.read_image(mapbdo.get("filepath"))
	# if mapemdata.get_ndim() != 3:
	# 	raise Exception, "Must be volume map"
	# mapheader = mapemdata.get_attr_dict()


	# XML: emdEntry / map / file
	e_file = etree.Element("file", format="MRC", sizeKb=sizeKb, type="map")
	e_map.append(e_file)


	####################
	# XML: emdEntry / supplement
	e_supplement = e.find("/supplement/")

	# XML: emdEntry / supplement / masks
	e_maskSet = etree.Element("maskSet")
	maskset = db.getchildren(root.recid, rectype=["mask"])
	for mask in maskset:
		maskbdos = db.getbinary([mask])
		for maskbdo in maskbdos:
			maskinfo = os.stat(maskbdo.get("filepath"))
			sizeKb = str(maskinfo.st_size / 1024)

			e_file = etree.Element("file", format="MRC", sizeKb=sizeKb, type="map")
			e_maskSet.append(e_file)

			e_details = etree.Element("details")
			e_details.text = db.renderview(mask, viewtype="recname")
			e_maskSet.append(e_details)


	e_sliceSet = etree.Element("sliceSet")
	e_figureSet = etree.Element("figureSet")
	e_fscSet = etree.Element("fscSet")

	e_supplement.append(e_maskSet)
	e_supplement.append(e_sliceSet)
	e_supplement.append(e_figureSet)
	e_supplement.append(e_fscSet)



	####################
	# XML: emdEntry / sample
	e_sample = e.find("/sample/")

	molWtTheo = etree.Element("molWtTheo")
	molWtTheo.text = str(subproject.get("mass_specimen") / 1000)
	e_sample.append(molWtTheo)

	e_name = etree.Element("name")
	e_name.text = subproject.get("name_specimen")

	e_sampleComponentList = etree.Element("sampleComponentList")

	samplelist = [None]
	for sample in samplelist:
		e_sampleComponent = etree.Element("sampleComponent", componentID="1")
		e_protein = etree.Element("protein")

		e_refInterpro = etree.Element("refInterpro")
		e_refInterpro.text = ""
		e_refGo = etree.Element("refGo")
		e_refGo.text = ""
		e_protein.append(e_refGo)

		e_mutantFlag = etree.Element("mutantFlag")
		e_mutantFlag.text = ""
		e_protein.append(e_mutantFlag)

		e_oligomericDetails = etree.Element("oligomericDetails")
		e_oligomericDetails.text = ""
		e_protein.append(e_oligomericDetails)

		e_numCopies = etree.Element("numCopies")
		e_numCopies.text = ""
		e_protein.append(e_numCopies)

		e_natSource = etree.Element("natSource")
		e_cell = etree.Element("cell")
		e_cell.text = ""
		e_natSource.append(e_cell)
		e_protein.append(e_natSource)

		e_engSource = etree.Element("engSource")
		e_vector = etree.Element("vector")
		e_expSystem = etree.Element("expSystem")
		e_engSource.append(e_vector)
		e_engSource.append(e_expSystem)
		e_protein.append(e_engSource)

		e_synName = etree.Element("synName")
		e_synName.text = ""
		e_protein.append(e_synName)

		e_entry = etree.Element("entry")
		e_entry.text = ""
		e_protein.append(e_entry)

		e_sciName = etree.Element("sciName")
		e_sciName.text = ""
		e_protein.append(e_sciName)

		e_molWtTheo = etree.Element("molWtTheo")
		e_molWtTheo.text = ""
		e_protein.append(e_molWtTheo)

		e_sampleComponent.append(e_protein)
		e_sample.append(e_sampleComponent)

	e_sample.append(e_sampleComponentList)



	####################
	# XML: emdEntry / experiment
	e_experiment = e.find("/experiment/")

	ccds = db.getchildren(root.recid, rectype=["ccd"])
	ccdrecs = db.getrecord(ccds)

	scans = db.getchildren(root.recid, rectype=["scan"], recurse=2)
	scanrecs = db.getrecord(scans)

	micrographs = db.getparents(scans, rectype=["micrograph"])
	micrographrecs = db.getrecord(micrographs)

	if ccds and scans:
		print "WARNING!! SCANS AND CCDS!!"

	#freeze_types = ["freezing", "freezing_berriman", "freezing_gatanmkii", "freezing_manual", "freezing_pneumatic", "freezing_vitrobot"]
	freeze_types = ["freezing_vitrobot"]
	freezes = emen2.util.listops.combine(db.getparenttree(ccds | micrographs, rectype=freeze_types, recurse=2).values())
	freezerecs = db.getrecord(freezes)

	instrument = ", ".join(set(filter(None, [i.rectype for i in freezerecs])))
	blottimes = ", ".join(set(filter(None, [str(i.get("vitrobot_time_blot")) for i in freezerecs])))
	vitrobottemp = ", ".join(set(filter(None, [str(i.get("vitrobot_temp")+273) for i in freezerecs])))
	rh = ", ".join(set(filter(None, [str(i.get("vitrobot_humidity")) for i in freezerecs])))
	cryogen = ", ".join(set(filter(None, [str(i.get("cryogen")) for i in freezerecs])))


	# XML: emdEntry / experiment / vitrification
	e_vitrification = etree.Element("vitrification")

	e_method = etree.Element("method")
	e_method.text = "Blot for %s seconds"%blottimes
	e_vitrification.append(e_method)

	e_cryogenName = etree.Element("cryogenName")
	e_cryogenName.text = cryogen
	e_vitrification.append(e_cryogenName)

	e_humidity = etree.Element("humidity")
	e_humidity.text = rh
	e_vitrification.append(e_humidity)

	e_instrument = etree.Element("instrument")
	e_instrument.text = instrument
	e_vitrification.append(e_instrument)

	e_temperature = etree.Element("temperature", units="Kelvin")
	e_temperature.text = vitrobottemp
	e_vitrification.append(e_temperature)

	e_experiment.append(e_vitrification)


	# XML: emdEntry / experiment / imaging
	e_imaging = etree.Element("imaging")

	microscopes = emen2.util.listops.combine(db.getparenttree(ccds | micrographs, recurse=4, rectype=["microscope"]).values())
	microscoperecs = db.getrecord(microscopes)

	grid_imagings = emen2.util.listops.combine(db.getparents(ccds | micrographs, recurse=4, rectype=["grid_imaging"]).values())
	grid_imaging_recs = db.getrecord(grid_imagings)

	microscopys = emen2.util.listops.combine(db.getparents(ccds | micrographs, recurse=2, rectype="microscopy").values())
	microscopyrecs = db.getrecord(microscopys)

	mags = set([i.get("tem_magnification_set") for i in ccdrecs+micrographrecs])
	print "mags"
	print mags

	defocuses = set([i.get("ctf_defocus_set") for i in ccdrecs+micrographrecs])
	print "defocuses"
	print defocuses

	doses = set([i.get("tem_dose_rate") for i in ccdrecs+micrographrecs])
	print "doses"
	print doses

	temperatures = set([i.get("temperature_specimen") for i in grid_imaging_recs])
	print "temperatures"
	print temperatures

	detectors = set([i.get("ccd_id") for i in ccdrecs]+[i.get("film_type") for i in microscopyrecs])
	print "detectors"
	print detectors

	cses = set([i.get("aberration_spherical") for i in microscoperecs])
	print "cses"
	print cses

	microscopenames = set([i.get("tem_name") for i in microscoperecs])
	print "microscopenames"
	print microscopenames

	dates = set([i.get("date_occurred") for i in ccdrecs+micrographrecs])
	print "dates"
	print dates

	voltages = set([i.get("lens_voltage_objective") for i in microscoperecs])
	print "voltages"
	print voltages

	e_electronSource = etree.Element("electronSource")
	e_electronSource.text = "FIELD EMISSION GUN"
	e_imaging.append(e_electronSource)

	e_electronDose = etree.Element("electronDose", units="e/A**2")
	e_electronDose.text = str(min(doses))
	e_imaging.append(e_electronDose)

	e_nominalDefocusMin = etree.Element("nominalDefocusMin", units="nM")
	e_nominalDefocusMin.text = str(min(defocuses)*1000)
	e_imaging.append(e_nominalDefocusMin)

	e_nominalDefocusMax = etree.Element("nominalDefocusMax", units="nM")
	e_nominalDefocusMax.text = str(max(defocuses)*1000)
	e_imaging.append(e_nominalDefocusMax)

	e_temperatureMin = etree.Element("temperatureMin", units="Kelvin")
	e_temperatureMin.text = str(min(temperatures))
	e_imaging.append(e_temperatureMin)

	e_temperatureMax = etree.Element("temperatureMax", units="Kelvin")
	e_temperatureMax.text = str(max(temperatures))
	e_imaging.append(e_temperatureMax)

	e_illuminationMode = etree.Element("illuminationMode")
	e_illuminationMode.text = "FLOOD BEAM"
	e_imaging.append(e_illuminationMode)

	e_specimenHolder = etree.Element("specimenHolder")
	e_specimenHolder.text = "Top entry"
	e_imaging.append(e_specimenHolder)

	e_detector = etree.Element("detector")
	e_detector.text = detectors.pop()
	e_imaging.append(e_detector)

	e_nominalCs = etree.Element("nominalCs")
	e_nominalCs.text = str(min(cses))
	e_imaging.append(e_nominalCs)

	e_temperature = etree.Element("temperature")
	e_temperature.text = str(max(temperatures))
	e_imaging.append(e_temperature)

	e_microscope = etree.Element("microscope")
	e_microscope.text = microscopenames.pop()
	e_imaging.append(e_microscope)

	e_date = etree.Element("date")
	e_date.text = ", ".join(dates)
	e_imaging.append(e_date)

	e_specimenHolderModel = etree.Element("specimenHolderModel")
	e_specimenHolderModel.text = "OTHER"
	e_imaging.append(e_specimenHolderModel)

	e_acceleratingVoltage = etree.Element("acceleratingVoltage", units="kV")
	e_acceleratingVoltage.text = str(max(voltages))
	e_imaging.append(e_acceleratingVoltage)

	e_nominalMagnification = etree.Element("nominalMagnification")
	e_nominalMagnification.text = str(max(mags)*1000)
	e_imaging.append(e_nominalMagnification)

	e_imagingMode = etree.Element("imagingMode")
	e_imagingMode.text = "OTHER"
	e_imaging.append(e_imagingMode)

	e_details = etree.Element("details")
	e_details.text = "low dose on %s"%e_microscope.text
	e_imaging.append(e_details)


	e_experiment.append(e_imaging)


	# XML: emdEntry / experiment / imageScans


	# XML: emdEntry / experiment / fitting


	# XML: emdEntry / experiment / samplePreparation




	####################
	# XML: emdEntry / processing

	# XML: emdEntry / processing / reconstruction

	# XML: emdEntry / processing / emDataSet





	# XML: done.. print
	# pretty_print(etree.tostring(e.find("/")))

	pretty = etree.tostring(e.find("/"))
	pr2 = xml.dom.minidom.parseString(pretty)
	return pr2.toprettyxml()

__version__ = "$Revision: 1.12 $".split(":")[1][:-1].strip()
